import { Actor } from './actor.js';
import { Graph } from './graph.js';
import { Action } from './action.js';
import { CryptoUtil } from './crypto.js';
import params from './params.js';

class Post extends Actor {
  constructor(forum, authorPublicKey, postContent) {
    const index = forum.posts.countVertices();
    const name = `Post${index + 1}`;
    super(name, forum.scene);
    this.id = postContent.id ?? CryptoUtil.randomUUID();
    this.authorPublicKey = authorPublicKey;
    this.value = 0;
    this.citations = postContent.citations;
    this.title = postContent.title;
    this.totalCitationWeight = this.citations.reduce((total, { weight }) => total += weight, 0);
    if (this.totalCitationWeight > params.revaluationLimit) {
      throw new Error('Post total citation weight exceeds revaluation limit '
      + `(${this.totalCitationWeight} > ${params.revaluationLimit})`);
    }
    if (this.citations.some(({ weight }) => Math.abs(weight) > 1)) {
      throw new Error('Each citation weight must be in the range [-1, 1]');
    }
  }

  async setPostValue(value) {
    await this.setValue('value', value);
    this.value = value;
  }

  getPostValue() {
    return this.value;
  }
}

/**
 * Purpose: Maintain a directed, acyclic, weighted graph of posts referencing other posts
 */
export class Forum extends Actor {
  constructor(name, scene) {
    super(name, scene);
    this.posts = new Graph(scene);
    this.actions = {
      addPost: new Action('add post', scene),
    };
  }

  async addPost(authorId, postContent) {
    const post = new Post(this, authorId, postContent);
    await this.actions.addPost.log(this, post);
    this.posts.addVertex(post.id, post, post.title);
    for (const { postId: citedPostId, weight } of post.citations) {
      this.posts.addEdge('citation', post.id, citedPostId, { weight });
    }
    return post.id;
  }

  getPost(postId) {
    return this.posts.getVertexData(postId);
  }

  getPosts() {
    return this.posts.getVertices();
  }

  async propagateValue(postId, increment, depth = 0) {
    if (depth > params.maxPropagationDepth) {
      return [];
    }
    const post = this.getPost(postId);
    const rewards = new Map();
    const addReward = (id, value) => rewards.set(id, (rewards.get(id) ?? 0) + value);
    const addRewards = (r) => {
      for (const [id, value] of r) {
        addReward(id, value);
      }
    };

    // Increment the value of the given post
    const postValue = post.getPostValue();
    await post.setPostValue(postValue + increment);

    // Award reputation to post author
    console.log('reward for post author', post.authorPublicKey, increment);
    addReward(post.authorPublicKey, increment);

    // Recursively distribute reputation to citations, according to weights
    for (const { postId: citedPostId, weight } of post.citations) {
      addRewards(await this.propagateValue(citedPostId, weight * increment, depth + 1));
    }

    return rewards;
  }
}
